
#ifndef __VDIF_HEADER__
#define __VDIF_HEADER__
/** VDIF Class Header 
 * @Version $Rev: 56 $
 * @Author M.Sekido
 * @Date  $Date: 2013-07-03 20:40:14 +0900 (Wed, 03 Jul 2013) $
 * $Id: VDIF_Header.hpp 56 2013-07-03 11:40:14Z vlbi $ 
 */

//#include "VLBI_Header.hpp"
#include  <iostream>
#include  <sstream>
#include  <ctime>
#include  <cstring>
#include  <cstdio>
#include  <cstdlib>
#include "time_lib.hpp"
using namespace std;

extern "C" {
#include  <stdint.h>
#include  <error.h>
#include  <errno.h>
}

const int VDIF_Hdr_size = 32;

//========== VDIF ===from Here  ====================

/// VDIF Sync patter of NICT Version
const uint32_t VDIF_NICT_SYNC = 0xACABFEED;

enum sample_rate_t {KSPS, MSPS};
/// VDIF Extended user data structure
struct VDIF_EUD_nict1{
  unsigned srate:23;    ///< Sampling Rate (ksps/Msps)
  unsigned sr_unit:1;   ///< Sampling Rate unit (0:ksps,1:Msps)
  uint8_t ext_dat_ver;  ///< Extended User data version Number
  uint32_t sync;        ///< Sync pattern 'ACABFEED'
  char    DAST_name[8]; ///< DAS/Station name 
  /// Default Constructor of VDIF_EUD_nict1
public:
  /// Constructor 
  VDIF_EUD_nict1(): srate(32000),sr_unit(KSPS),ext_dat_ver(1), 
		    sync(VDIF_NICT_SYNC){
    memset(DAST_name,0,8);}
  /// Initialize the header
  void Init(const char* name="KASHIM34"){
    ext_dat_ver=1; sync=VDIF_NICT_SYNC; 
    strncpy(DAST_name,name,8);
  }
  /// Set Sampling Rate (kHz/MHz)
  void Set_Srate(unsigned long sr, const sample_rate_t Msps_unit=KSPS){
    srate=sr; sr_unit=Msps_unit;}
  /// Set Station name
  void Set_stname(const char *name){strncpy(DAST_name,name,8);}
  /// Returns Station name in string 
  void Get_Stname(string& name)const;
  /// Returns Extended User Data Version 
  int ExDataVer()const {return ext_dat_ver;}
};

inline void VDIF_EUD_nict1::Get_Stname(string& name)const{
  char buf[10]; 
  strncpy(buf,DAST_name,8); buf[8]='\0';
  name = buf;
}

struct VDIF_Hdr{
/** Seconds from reference Epoch **/
    unsigned  Sec_from_Ref:30;
    /** Flag for legacy compatibility, 0:32 Bytes header, 1:16 Bytes header*/
    unsigned  Legacy_flag:1;
    /** Invalid flag 1: Invalid, 0: Valid */
    unsigned  Invalid_flag:1;
    /** Data Frame Number */
    unsigned  Frame_Number:24;
    /** Reference Epoch half year counter since 2000.0 */
    unsigned  Ref_Epoch_hy_counter:6;
    /** unused  */
    unsigned  unused:2;
      /** Data Frame length -1, unit: 8 bytes  */
      unsigned  Frame_Length:24;
      /** Log2(nch)*/
      unsigned  Lg_Nch:5;
      /** VDIF Version Number */
      unsigned  Version_No:3;
    /** Station ID*/
    unsigned char Station_ID[2];
    /** Thread ID*/
    unsigned  Thread_Id:10;
  /** Quantization Bit/sample -1 */
    unsigned  Qbit_1:5;
    /** Comlex data flag. 1:Complex, 0: Real data */
    unsigned  Complex_data_flag:1;
  /** Extended user data 1 */
  VDIF_EUD_nict1 exud_nict;
public:
  /// Constructor 
  VDIF_Hdr(){
    memset(this,0,sizeof(VDIF_Hdr)); 
    Version_No=1;Thread_Id=0;Qbit_1=1; 
    strncpy((char*)Station_ID,"kb",2);
    exud_nict.Init();
  }
  ~VDIF_Hdr(){;  }  ///< Destructor 
  const bool Valid() const { return ( Invalid_flag ? false : true);}
  /// Returns Total second from Reference Epoch 
  uint32_t Sec_Ref()const{ return (uint32_t)Sec_from_Ref;}
  int32_t  iFrame()const{return Frame_Number;} ///< Returns Frame Number 
  int Nch() const {return (1 << Lg_Nch); }  ///< Returns Nch
  int Qbit()const {return Qbit_1+1;}  ///< returns Qunatization bits
  /// Sampling rate (Hz)
  uint32_t Srate()const{return Srate_ksps()*1000;}
  //  time_t Ref_Epoch() const;  ///< Reference Epoch
  Time_dat Ref_Epoch() const;  ///< Reference Epoch
  /**! Printing Header data
   * \param fmt pattern 1,2,20,21
   */
  void Print_hd(FILE *fp=stdout, const int fmt=1){
    fprintf(fp,"%s\n", Str_hd(fmt).c_str());
  };
  /**! Header data printing to string 
   * \param fmt pattern 1,2,20,21
   * \return string 
   */
  string Str_hd(const int fmt=1);
  /// return framesize including header (bytes)
  int Framesize(){long fs = (Frame_Length)*8; 
    return ( fs < VDIF_Hdr_size ? VDIF_Hdr_size : fs); }
  /**!
   * \return header size (bytes)
   */
  int Headsize()const {return (VDIF_Hdr_size);}
  /**!
   * \return Time_dat data of the epoch
   */
  Time_dat Epoch()const;
  /// Return integer MJD of the epoch
  uint32_t Epoch_imjd(){return (uint32_t)(floor(Epoch().Mjd()));}
  /// Returns Second of the day
  double Epoch_sod() {return Epoch().Sod();}  ///< return sec of the day (sec)  
  /// Return header type
  //  VLBI_Hdr_t Header_type()const {return VDIF_Header_Type;}
  bool HD_Match() const{return (exud_nict.sync == VDIF_NICT_SYNC ? true: false);}
  /// Setting  Nch 
  void Set_Nch(const int nch);
  /// Setting Station ID
  void Set_Stnid(const char *id){strncpy((char*)Station_ID,(char*)id,2);}
  /// Setting Quantization bit
  void Set_Qbit(const int nbit){Qbit_1 = nbit-1;}
  /// Setting Sampling Freq (ksps);
  void Set_Srate(const uint32_t ksps);
  /// Return Sampling rate in ksps
  uint32_t Srate_ksps()const; 
  /// Setting Framesize (Bytes)
  void Set_FrameLen(const uint32_t fbyte){Frame_Length = fbyte/8;}
  /// Setting Sampling Parameters and Framesize at a time
  int  Set_Sparam(const int nch, const int qbit, const uint32_t srate);
  /// Setup Epoch parametes;
  void Set_Epoch(const Time_dat &t1);
  /// Set Initial Epoch. 
  /// When t1 is given, nearst ealier Epoch is automatically set.
  Time_dat Set_Epoch0(const time_t t1);
  //  void Set_Epoch(const time_t t1, const time_t t0_ref);
  /// Set Frame number 
  void Set_iFrame(const uint32_t iframe){ Frame_Number = iframe;}
  /// Counting up Frame Number ;
  void iFrame_plus(){Frame_Number++;}
  /// Returns Total Data rate in word(4 bytes)/sec
  int32_t Speed()const { 
    if(HD_Match()) return (Srate()/32 * Nch() * Qbit()); else return -1;
  }
  /// Returns Data Type (Real/Complex)
  string Dtype(void)const {string ret = (Complex_data_flag == 1 ? "Cmplx" :"Real"); 
    return ret;}
  /// Get Station ID
  void Get_Stnid(string & id) const{ 
    char buf[3]; strncpy(buf,(char*)Station_ID,2); buf[2]='\0'; id=buf;}
  const string Stnid(void) const{ 
    char buf[5]; strncpy(buf,(char*)Station_ID,2); buf[2]='\0';
    string ss(buf);
    return ss; }
  const string Stname() const{
    string str="N/A"; 
    if(exud_nict.ext_dat_ver == 1)  exud_nict.Get_Stname(str);
    return str; }
  const string Version(void){
    char str[256];
    sprintf(str, "Ver.%d.%d", Version_No,exud_nict.ext_dat_ver);
    string ss = str;
    return ss;
  }
  const int ThreadID(void) {
    int thid = Thread_Id;
    return thid;
  }
};

inline  void VDIF_Hdr::Set_Srate(const uint32_t ksps){
  if(ksps <1000000){
    exud_nict.Set_Srate(ksps, KSPS);
  }else{
    exud_nict.Set_Srate(ksps/1000000,MSPS);
  }
}

inline  uint32_t VDIF_Hdr::Srate_ksps()const{
  if(exud_nict.sr_unit){ // Msps
    return exud_nict.srate*1000;
  }else{ // ksps
    return exud_nict.srate;
  }
}

inline  void VDIF_Hdr::Set_Nch(const int nch){
  int lch =0;
  int ich=1;
  while( (nch & ich) == 0){ ich <<= 1; lch++;}
  Lg_Nch = lch;
}

inline Time_dat VDIF_Hdr::Epoch()const{
  Time_dat t0;
  t0 =  Ref_Epoch();
  t0 += Sec_from_Ref;
  return t0;
}


enum VDIF_Ext_t {No_ext, NICT_ext };

const bool VDIF_Hdr_find(const uint8_t*  buf, const uint32_t nbyte,
			 long &offset, VDIF_Hdr* hdr);

const bool VDIF_Hdr_find(FILE *fp, long &offset, VDIF_Hdr* hdr);

// /// @class VDIF_Header is header class of VDIF
// class VDIF_Header: public VLBI_Header{
//   VDIF_Hdr* pt_head;
// public:
//   VDIF_Header():VLBI_Header(VDIF_Hdr_size),pt_head(0){;} /// < Default Constructor 
//   /// Header finder
//   /// @return  Number of bytes to find the header, or -1: for Error
//   int32_t Hd_find(const uint8_t*  buf, const uint32_t nbyte);
//   /// Print Header Information
// };

//========== VDIF ===to   Here  ====================

#endif
